import datetime as dt
from unittest.mock import Mock, patch
from urllib.parse import parse_qs, urlparse

from django.contrib.sessions.middleware import SessionMiddleware
from django.test import RequestFactory, TestCase
from django.urls import reverse
from django.utils.dateparse import parse_datetime
from django.utils.timezone import now

from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import create_user_from_evecharacter, json_response_to_python

from structures import views
from structures.constants import EveTypeId
from structures.models import Owner, PocoDetails, Structure, StructureItem, Webhook

from .testdata.factories import create_owner_from_user, create_poco, create_starbase
from .testdata.helpers import (
    create_structures,
    load_entities,
    load_entity,
    set_owner_character,
)
from .testdata.load_eveuniverse import load_eveuniverse

VIEWS_PATH = "structures.views"
OWNERS_PATH = "structures.models.owners"


def json_response_to_dict(response, key="id") -> dict:
    """Convert JSON response into dict by given key."""
    return {x[key]: x for x in json_response_to_python(response)["data"]}


class TestStructureListFormat(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        create_structures()

    def test_should_format_rows_correctly(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_structures", user
        )
        structure = Structure.objects.get(id=1000000000001)
        # when
        request = self.factory.get(reverse("structures:structure_list_data"))
        request.user = user
        response = views.structure_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        obj = data[1000000000001]
        self.assertEqual(obj["alliance_name"], "Wayne Enterprises")
        self.assertEqual(obj["corporation_name"], "Wayne Technologies")
        self.assertEqual(obj["region_name"], "Heimatar")
        self.assertEqual(obj["solar_system_name"], "Amamake")
        self.assertEqual(obj["group_name"], "Citadel")
        self.assertEqual(obj["category_name"], "Structure")
        self.assertFalse(obj["is_starbase"])
        self.assertFalse(obj["is_poco"])
        self.assertEqual(obj["type_name"], "Astrahus")
        self.assertEqual(obj["is_reinforced_str"], "no")
        self.assertEqual(obj["reinforcement"], "05:00")
        self.assertEqual(
            obj["fuel_expires_at"]["timestamp"], structure.fuel_expires_at.isoformat()
        )
        self.assertEqual(obj["power_mode_str"], "Full Power")
        self.assertEqual(obj["state_str"], "Shield vulnerable")
        self.assertEqual(obj["core_status_str"], "")
        self.assertEqual(obj["details"], "")


class TestStructureList(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        create_structures()

    def test_should_have_access_to_main_view(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        # when
        request = self.factory.get(reverse("structures:main"))
        request.user = user
        response = views.main(request)
        # then
        self.assertEqual(response.status_code, 200)

    # TODO: This tests fails on sqlite, re-enable it with mysql
    # def test_should_return_summary_data(self):
    #     # given
    #     user, _ = set_owner_character(character_id=1001)
    #     user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
    #     # when
    #     request = self.factory.get(reverse("structures:structure_summary_data"))
    #     request.user = user
    #     response = views.structure_summary_data(request)
    #     # then
    #     self.assertEqual(response.status_code, 200)
    #     data = json_response_to_dict(response)
    #     obj = data[2001]
    #     self.assertEqual(obj["corporation_name"], "Wayne Technologies")
    #     self.assertEqual(obj["alliance_name"], "Wayne Enterprises")
    #     self.assertEqual(obj["citadel_count"], 1)
    #     self.assertEqual(obj["ec_count"], 0)
    #     self.assertEqual(obj["refinery_count"], 1)
    #     self.assertEqual(obj["other_count"], 0)
    #     self.assertEqual(obj["poco_count"], 4)
    #     self.assertEqual(obj["starbase_count"], 3)
    #     self.assertEqual(obj["total"], 9)


class TestStructureListSpecial(TestCase):
    def test_should_show_empty_list(self):
        # given
        load_entity(EveCharacter)
        user, _ = create_user_from_evecharacter(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        # when
        request = RequestFactory().get(reverse("structures:main"))
        request.user = user
        response = views.main(request)
        # then
        self.assertEqual(response.status_code, 200)


class TestStructureListDataPermissions(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        create_structures()

    def _structure_list_data_view(self, user) -> dict:
        """helper method:  makes the request to the view
        and returns response as dict for the given user
        """
        request = self.factory.get(reverse("structures:structure_list_data"))
        request.user = user
        response = views.structure_list_data(request)
        self.assertEqual(response.status_code, 200)
        return json_response_to_dict(response)

    def test_should_show_no_structures(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        # when
        structure_ids = self._structure_list_data_view(user).keys()
        # then
        self.assertSetEqual(set(structure_ids), set())

    def test_should_show_own_corporation_only_1(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_corporation_structures", user
        )
        # when
        structure_ids = self._structure_list_data_view(user).keys()
        # then
        self.assertSetEqual(
            set(structure_ids),
            {
                1000000000001,
                1000000000002,
                1200000000003,
                1200000000004,
                1200000000005,
                1200000000006,
                1300000000001,
                1300000000002,
                1300000000003,
            },
        )

    def test_should_show_own_corporation_only_2(self):
        # given
        user, _ = set_owner_character(character_id=1011)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_corporation_structures", user
        )
        # when
        structure_ids = self._structure_list_data_view(user).keys()
        # then
        self.assertSetEqual(set(structure_ids), {1000000000003, 1000000000004})

    def test_should_show_own_alliance_only_1(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_alliance_structures", user
        )
        # when
        structure_ids = self._structure_list_data_view(user).keys()
        # then
        self.assertSetEqual(
            set(structure_ids),
            {
                1000000000001,
                1000000000002,  # only for alliance
                1200000000003,
                1200000000004,
                1200000000005,
                1200000000006,
                1300000000001,
                1300000000002,
                1300000000003,
            },
        )

    def test_should_show_own_alliance_only_2(self):
        # given
        user, _ = set_owner_character(character_id=1011)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_alliance_structures", user
        )
        # when
        structure_ids = self._structure_list_data_view(user).keys()
        # then
        self.assertSetEqual(set(structure_ids), {1000000000003, 1000000000004})

    def test_should_show_all_structures(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_structures", user
        )
        # when
        structure_ids = self._structure_list_data_view(user).keys()
        # then
        self.assertSetEqual(
            set(structure_ids),
            {
                1000000000001,
                1000000000002,
                1000000000003,  # only for all
                1000000000004,
                1200000000003,
                1200000000004,
                1200000000005,
                1200000000006,
                1300000000001,
                1300000000002,
                1300000000003,
            },
        )

    def test_should_show_unanchoring_status(self):
        # given
        user, _ = set_owner_character(character_id=1011)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_corporation_structures", user
        )
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_unanchoring_status", user
        )
        # when
        data = self._structure_list_data_view(user)
        # then
        structure = data[1000000000003]
        self.assertIn("Unanchoring until", structure["state_details"])

    def test_should_not_show_unanchoring_status(self):
        # given
        user, _ = set_owner_character(character_id=1011)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_corporation_structures", user
        )
        # when
        data = self._structure_list_data_view(user)
        # then
        structure = data[1000000000003]
        self.assertNotIn("Unanchoring until", structure["state_details"])


class TestStructureListFilters(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "structures.basic_access", cls.user
        )
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_structures", cls.user
        )
        cls.factory = RequestFactory()

    @patch(VIEWS_PATH + ".STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED", True)
    def test_default_filter_enabled(self):
        request = self.factory.get(reverse("structures:index"))
        request.user = self.user
        response = views.index(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/structures/list?tags=tag_a")

    @patch(VIEWS_PATH + ".STRUCTURES_DEFAULT_TAGS_FILTER_ENABLED", False)
    def test_default_filter_disabled(self):
        request = self.factory.get(reverse("structures:index"))
        request.user = self.user
        response = views.index(request)
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, "/structures/list")

    def test_list_filter_by_tag_1(self):
        # no filter
        request = self.factory.get(
            "{}".format(reverse("structures:structure_list_data"))
        )
        request.user = self.user
        response = views.structure_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json_response_to_dict(response)
        self.assertSetEqual(
            set(data.keys()),
            {
                1000000000001,
                1000000000002,
                1000000000003,
                1000000000004,
                1200000000003,
                1200000000004,
                1200000000005,
                1200000000006,
                1300000000001,
                1300000000002,
                1300000000003,
            },
        )

        # filter for tag_c
        request = self.factory.get(
            "{}?tags=tag_c".format(reverse("structures:structure_list_data"))
        )
        request.user = self.user
        response = views.structure_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json_response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {1000000000002, 1000000000003})

        # filter for tag_b
        request = self.factory.get(
            "{}?tags=tag_b".format(reverse("structures:structure_list_data"))
        )
        request.user = self.user
        response = views.structure_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json_response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {1000000000003})

        # filter for tag_c, tag_b
        request = self.factory.get(
            "{}?tags=tag_c,tag_b".format(reverse("structures:structure_list_data"))
        )
        request.user = self.user
        response = views.structure_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json_response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {1000000000002, 1000000000003})

    def test_call_with_raw_tags(self):
        request = self.factory.get(
            "{}?tags=tag_c,tag_b".format(reverse("structures:main"))
        )
        request.user = self.user
        response = views.main(request)
        self.assertEqual(response.status_code, 200)

    def test_set_tags_filter(self):
        request = self.factory.post(
            reverse("structures:main"),
            data={
                "tag_b": True,
                "tag_c": True,
            },
        )
        request.user = self.user
        response = views.main(request)
        self.assertEqual(response.status_code, 302)
        parts = urlparse(response.url)
        path = parts[2]
        query_dict = parse_qs(parts[4])
        self.assertEqual(path, reverse("structures:main"))
        self.assertIn("tags", query_dict)
        params = query_dict["tags"][0].split(",")
        self.assertSetEqual(set(params), {"tag_c", "tag_b"})


class TestStructurePowerModes(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        load_eveuniverse()

    def setUp(self):
        create_structures(dont_load_entities=True)
        self.user, self.owner = set_owner_character(character_id=1001)
        AuthUtils.add_permission_to_user_by_name("structures.basic_access", self.user)
        AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_structures", self.user
        )

    def display_data_for_structure(self, structure_id: int):
        request = self.factory.get(reverse("structures:structure_list_data"))
        request.user = self.user
        response = views.structure_list_data(request)
        self.assertEqual(response.status_code, 200)

        data = json_response_to_python(response)["data"]
        for row in data:
            if row["id"] == structure_id:
                return row

        return None

    def test_full_power(self):
        structure_id = 1000000000001
        structure = Structure.objects.get(id=structure_id)
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.save()
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "Full Power")
        self.assertEqual(
            parse_datetime(my_structure["fuel_expires_at"]["timestamp"]),
            structure.fuel_expires_at,
        )
        self.assertIn("Full Power", my_structure["last_online_at"]["display"])

    def test_low_power(self):
        structure_id = 1000000000001
        structure = Structure.objects.get(id=structure_id)
        structure.fuel_expires_at = None
        structure.last_online_at = now() - dt.timedelta(days=3)
        structure.save()
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "Low Power")
        self.assertEqual(
            parse_datetime(my_structure["last_online_at"]["timestamp"]),
            structure.last_online_at,
        )
        self.assertIn("Low Power", my_structure["fuel_expires_at"]["display"])

    def test_abandoned(self):
        structure_id = 1000000000001
        structure = Structure.objects.get(id=structure_id)
        structure.fuel_expires_at = None
        structure.last_online_at = now() - dt.timedelta(days=7, seconds=1)
        structure.save()
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "Abandoned")
        self.assertIn("Abandoned", my_structure["fuel_expires_at"]["display"])
        self.assertIn("Abandoned", my_structure["last_online_at"]["display"])

    def test_maybe_abandoned(self):
        structure_id = 1000000000001
        structure = Structure.objects.get(id=structure_id)
        structure.fuel_expires_at = None
        structure.last_online_at = None
        structure.save()
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "Abandoned?")
        self.assertIn("Abandoned?", my_structure["fuel_expires_at"]["display"])
        self.assertIn("Abandoned?", my_structure["last_online_at"]["display"])

    def test_poco(self):
        structure_id = 1200000000003
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "")
        self.assertIn("-", my_structure["fuel_expires_at"]["display"])
        self.assertIn("-", my_structure["last_online_at"]["display"])

    def test_starbase_online(self):
        structure_id = 1300000000001
        structure = Structure.objects.get(id=structure_id)
        structure.fuel_expires_at = now() + dt.timedelta(hours=1)
        structure.save()
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "")
        self.assertEqual(
            parse_datetime(my_structure["fuel_expires_at"]["timestamp"]),
            structure.fuel_expires_at,
        )
        self.assertIn("-", my_structure["last_online_at"]["display"])

    def test_starbase_offline(self):
        structure_id = 1300000000001
        structure = Structure.objects.get(id=structure_id)
        structure.fuel_expires_at = None
        structure.save()
        my_structure = self.display_data_for_structure(structure_id)
        self.assertEqual(my_structure["power_mode_str"], "")
        self.assertIn("-", my_structure["fuel_expires_at"]["display"])
        self.assertIn("-", my_structure["last_online_at"]["display"])


class TestAddStructureOwner(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities([EveCorporationInfo, EveAllianceInfo, EveCharacter, Webhook])
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
            permissions=["structures.basic_access", "structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        cls.character = cls.character_ownership.character
        cls.factory = RequestFactory()

    def _add_structure_owner(self, token=None, user=None):
        # given
        request = self.factory.get(reverse("structures:add_structure_owner"))
        if not user:
            user = self.user
        if not token:
            token = user.token_set.first()
        request.user = user
        request.token = token
        middleware = SessionMiddleware(Mock())
        middleware.process_request(request)
        orig_view = views.add_structure_owner.__wrapped__.__wrapped__.__wrapped__
        # when
        return orig_view(request, token)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", True)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_new_structure_owner_and_notify_admins(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # when
        response = self._add_structure_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertTrue(mock_notify_admins.called)
        owner = Owner.objects.first()
        self.assertSetEqual(
            {self.character_ownership.pk},
            set(owner.characters.values_list("character_ownership", flat=True)),
        )
        self.assertEqual(owner.webhooks.first().name, "Test Webhook 1")
        self.assertTrue(mock_update_all_for_owner.delay.called)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_character_to_existing_structure_owner_and_reactive(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # given
        owner = Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2102),
            is_active=False,
        )
        _, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner.add_character(character_ownership_1011)
        user_1102, character_ownership_1102 = create_user_from_evecharacter(
            1102,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        # when
        response = self._add_structure_owner(user=user_1102)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_update_all_for_owner.delay.called)
        owner.refresh_from_db()
        self.assertSetEqual(
            {character_ownership_1011.pk, character_ownership_1102.pk},
            set(owner.characters.values_list("character_ownership", flat=True)),
        )
        self.assertTrue(owner.is_active)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_new_structure_owner_and_not_notify_admins(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # when
        response = self._add_structure_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        owner = Owner.objects.first()
        self.assertSetEqual(
            {self.character_ownership.pk},
            set(owner.characters.values_list("character_ownership", flat=True)),
        )
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_notify_admins.called)
        self.assertTrue(mock_update_all_for_owner.delay.called)

    @patch(VIEWS_PATH + ".STRUCTURES_ADMIN_NOTIFICATIONS_ENABLED", False)
    @patch(VIEWS_PATH + ".tasks.update_all_for_owner")
    @patch(VIEWS_PATH + ".notify_admins")
    @patch(VIEWS_PATH + ".messages")
    def test_should_add_structure_owner_with_no_default_webhook(
        self, mock_messages, mock_notify_admins, mock_update_all_for_owner
    ):
        # given
        Webhook.objects.filter(name="Test Webhook 1").update(is_default=False)
        # when
        response = self._add_structure_owner()
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.info.called)
        self.assertFalse(mock_notify_admins.called)
        my_owner = Owner.objects.get(
            characters__character_ownership=self.character_ownership
        )
        self.assertIsNone(my_owner.webhooks.first())
        self.assertTrue(mock_update_all_for_owner.delay.called)

        # webhook.is_default = True
        # webhook.save()

    @patch(VIEWS_PATH + ".messages")
    def test_should_report_error_when_token_does_not_belong_to_user(
        self, mock_messages
    ):
        # given
        other_user, _ = create_user_from_evecharacter(
            1011,
            permissions=["structures.basic_access", "structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        # when
        my_token = other_user.token_set.first()
        response = self._add_structure_owner(token=my_token)
        # then
        self.assertEqual(response.status_code, 302)
        self.assertEqual(response.url, reverse("structures:index"))
        self.assertTrue(mock_messages.error.called)


class TestStatus(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        my_user, _ = set_owner_character(character_id=1001)
        AuthUtils.add_permission_to_user_by_name("structures.basic_access", my_user)
        cls.factory = RequestFactory()

    def test_view_service_status_ok(self):
        for owner in Owner.objects.filter(is_included_in_service_status=True):
            owner.structures_last_update_at = now()
            owner.notifications_last_update_at = now()
            owner.forwarding_last_update_at = now()
            owner.assets_last_update_at = now()
            owner.save()

        request = self.factory.get(reverse("structures:service_status"))
        response = views.service_status(request)
        self.assertEqual(response.status_code, 200)

    @patch(OWNERS_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    def test_view_service_status_fail_structures(self):
        for owner in Owner.objects.filter(is_included_in_service_status=True):
            owner.structures_last_update_at = now() - dt.timedelta(minutes=31)
            owner.notifications_last_update_at = now()
            owner.forwarding_last_update_at = now()
            owner.assets_last_update_at = now()
            owner.save()

        request = self.factory.get(reverse("structures:service_status"))
        response = views.service_status(request)
        self.assertEqual(response.status_code, 500)

    @patch(OWNERS_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_view_service_status_fail_notifications(self):
        for owner in Owner.objects.filter(is_included_in_service_status=True):
            owner.structures_last_update_at = now()
            owner.notifications_last_update_at = now() - dt.timedelta(minutes=31)
            owner.forwarding_last_update_at = now()
            owner.assets_last_update_at = now()
            owner.save()

        request = self.factory.get(reverse("structures:service_status"))
        response = views.service_status(request)
        self.assertEqual(response.status_code, 500)

    @patch(OWNERS_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_view_service_status_fail_forwarding(self):
        for owner in Owner.objects.filter(is_included_in_service_status=True):
            owner.structures_last_update_at = now()
            owner.notifications_last_update_at = now()
            owner.forwarding_last_update_at = now() - dt.timedelta(minutes=31)
            owner.assets_last_update_at = now()
            owner.save()

        request = self.factory.get(reverse("structures:service_status"))
        response = views.service_status(request)
        self.assertEqual(response.status_code, 500)

    @patch(OWNERS_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    def test_view_service_status_fail_assets(self):
        for owner in Owner.objects.filter(is_included_in_service_status=True):
            owner.structures_last_update_at = now()
            owner.notifications_last_update_at = now()
            owner.forwarding_last_update_at = now()
            owner.assets_last_update_at = now() - dt.timedelta(minutes=31)
            owner.save()

        request = self.factory.get(reverse("structures:service_status"))
        response = views.service_status(request)
        self.assertEqual(response.status_code, 500)


class TestPocoList(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "structures.basic_access", cls.user
        )
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_structures", cls.user
        )
        cls.factory = RequestFactory()

    def test_should_return_correct_data_for_poco(self):
        # given
        request = self.factory.get(reverse("structures:poco_list_data"))
        request.user = self.user
        self.owner.are_pocos_public = True
        self.owner.save()
        PocoDetails.objects.create(
            structure_id=1200000000003,
            alliance_tax_rate=0.02,
            allow_access_with_standings=True,
            allow_alliance_access=True,
            corporation_tax_rate=0.01,
            reinforce_exit_end=21,
            reinforce_exit_start=18,
        )
        # when
        response = views.poco_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        obj = data[1200000000003]
        self.assertEqual(obj["region"], "Heimatar")
        self.assertEqual(obj["solar_system"], "Amamake")
        self.assertEqual(obj["planet"], "Amamake V")
        self.assertEqual(obj["planet_type_name"], "Barren")
        self.assertEqual(obj["space_type"], "lowsec")
        self.assertEqual(obj["has_access_str"], "yes")
        self.assertEqual(obj["tax"], "1 %")

    def test_should_return_all_pocos(self):
        # given
        request = self.factory.get(reverse("structures:poco_list_data"))
        request.user = self.user
        self.owner.are_pocos_public = True
        self.owner.save()
        # when
        response = views.poco_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertSetEqual(
            set(data.keys()),
            {1200000000003, 1200000000004, 1200000000005, 1200000000006},
        )

    def test_should_return_no_pocos(self):
        # given
        request = self.factory.get(reverse("structures:poco_list_data"))
        request.user = self.user
        self.owner.are_pocos_public = False
        self.owner.save()
        # when
        response = views.poco_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertFalse(data)


class TestStructureFittingModal(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        create_structures()

    def test_should_have_access_to_fitting(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_structure_fit", user
        )
        # when
        request = self.factory.get(
            reverse("structures:structure_details", args=[1000000000001])
        )
        request.user = user
        response = views.structure_details(request, 1000000000001)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_not_have_access_to_fitting(self):
        # given
        user, _ = set_owner_character(character_id=1001)
        user = AuthUtils.add_permission_to_user_by_name("structures.basic_access", user)
        # when
        request = self.factory.get(
            reverse("structures:structure_details", args=[1000000000001])
        )
        request.user = user
        response = views.structure_details(request, 1000000000001)
        # then
        self.assertEqual(response.status_code, 302)


class TestDetailsModal(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_eveuniverse()
        load_entities()
        cls.user, _ = create_user_from_evecharacter(
            1001, permissions=["structures.basic_access"]
        )
        cls.owner = create_owner_from_user(cls.user)

    def test_should_load_poco_detail(self):
        # given
        structure = create_poco(owner=self.owner)
        # when
        request = self.factory.get(
            reverse("structures:poco_details", args=[structure.id])
        )
        request.user = self.user
        response = views.poco_details(request, structure.id)
        # then
        self.assertEqual(response.status_code, 200)

    def test_should_load_starbase_detail(self):
        # given
        structure = create_starbase(owner=self.owner)
        # when
        request = self.factory.get(
            reverse("structures:starbase_detail", args=[structure.id])
        )
        request.user = self.user
        response = views.starbase_detail(request, structure.id)
        # then
        self.assertEqual(response.status_code, 200)


class TestJumpGateList(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "structures.basic_access", cls.user
        )
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "structures.view_all_structures", cls.user
        )
        cls.factory = RequestFactory()

    def test_should_return_jump_gates(self):
        # given
        request = self.factory.get(reverse("structures:jump_gates_list_data"))
        request.user = self.user
        structure = Structure.objects.get(id=1000000000004)
        structure.items.create(
            id=1,
            eve_type_id=EveTypeId.LIQUID_OZONE,
            location_flag=StructureItem.LocationFlag.STRUCTURE_FUEL,
            is_singleton=False,
            quantity=5000,
        )
        # when
        response = views.jump_gates_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        data = json_response_to_dict(response)
        self.assertSetEqual(set(data.keys()), {1000000000004})
        obj = data[1000000000004]
        self.assertEqual(obj["region_name"], "Detorid")
        self.assertEqual(obj["solar_system_name"], "1-PGSG")
        self.assertEqual(
            obj["structure_name"], "1-PGSG &gt;&gt; A-C5TC - Test Jump Gate"
        )
        self.assertEqual(obj["jump_fuel_quantity"], 5000)



import datetime as dt
from unittest.mock import patch

from django.utils.timezone import now, utc
from eveuniverse.models import EvePlanet

from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from structures.core.notification_types import NotificationType
from structures.models import (
    FuelAlertConfig,
    Owner,
    PocoDetails,
    StarbaseDetail,
    Structure,
    StructureService,
    StructureTag,
    Webhook,
)
from structures.tests import to_json
from structures.tests.testdata.factories import (
    create_owner_from_user,
    create_poco,
    create_starbase,
    create_structure_service,
    create_upwell_structure,
)
from structures.tests.testdata.helpers import load_entities
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.owners"


@patch(MODULE_PATH + ".esi")
class TestUpdateStructuresEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # given (global)
        load_eveuniverse()
        load_entities()
        cls.user, _ = create_user_from_evecharacter(
            1001,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        Webhook.objects.all().delete()
        cls.endpoints = [
            EsiEndpoint(
                "Assets",
                "get_corporations_corporation_id_assets",
                "corporation_id",
                needs_token=True,
                data={
                    "2001": [
                        {
                            "is_singleton": False,
                            "item_id": 1300000001001,
                            "location_flag": "QuantumCoreRoom",
                            "location_id": 1000000000001,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 56201,
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1300000001002,
                            "location_flag": "ServiceSlot0",
                            "location_id": 1000000000001,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 35894,
                        },
                        {
                            "is_singleton": True,
                            "item_id": 1300000002001,
                            "location_flag": "ServiceSlot0",
                            "location_id": 1000000000002,
                            "location_type": "item",
                            "quantity": 1,
                            "type_id": 35894,
                        },
                    ],
                    "2102": [
                        {
                            "is_singleton": False,
                            "item_id": 1300000003001,
                            "location_flag": "StructureFuel",
                            "location_id": 1000000000004,
                            "location_type": "item",
                            "quantity": 5000,
                            "type_id": 16273,
                        }
                    ],
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_locations",
                "corporation_id",
                needs_token=True,
                data={
                    "2001": [
                        {
                            "item_id": 1200000000003,
                            "position": {"x": 1.2, "y": 2.3, "z": -3.4},
                        },
                        {
                            "item_id": 1200000000004,
                            "position": {"x": 5.2, "y": 6.3, "z": -7.4},
                        },
                        {
                            "item_id": 1200000000005,
                            "position": {"x": 1.2, "y": 6.3, "z": -7.4},
                        },
                        {
                            "item_id": 1200000000006,
                            "position": {"x": 41.2, "y": 26.3, "z": -47.4},
                        },
                        {
                            "item_id": 1300000000001,
                            "position": {"x": 40.2, "y": 27.3, "z": -19.4},
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Assets",
                "post_corporations_corporation_id_assets_names",
                "corporation_id",
                needs_token=True,
                data={
                    "2001": [
                        {
                            "item_id": 1200000000003,
                            "name": "Customs Office (Amamake V)",
                        },
                        {
                            "item_id": 1200000000004,
                            "name": "Customs Office (1-PGSG VI)",
                        },
                        {
                            "item_id": 1200000000005,
                            "name": "Customs Office (1-PGSG VII)",
                        },
                        {
                            "item_id": 1200000000006,
                            "name": '<localized hint="Customs Office">Customs Office*</localized> (1-PGSG VIII)',
                        },
                        {"item_id": 1300000000001, "name": "Home Sweat Home"},
                        {"item_id": 1300000000002, "name": "Bat cave"},
                        {"item_id": 1300000000003, "name": "Panic Room"},
                    ]
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_structures",
                "corporation_id",
                needs_token=True,
                data={
                    "2001": [
                        {
                            "corporation_id": 2001,
                            "fuel_expires": dt.datetime(2020, 3, 5, 5, tzinfo=utc),
                            "next_reinforce_apply": None,
                            "next_reinforce_hour": None,
                            "profile_id": 52436,
                            "reinforce_hour": 19,
                            "services": [
                                {"name": "Reprocessing", "state": "online"},
                                {"name": "Moon Drilling", "state": "online"},
                            ],
                            "state": "shield_vulnerable",
                            "state_timer_end": None,
                            "state_timer_start": None,
                            "structure_id": 1000000000002,
                            "system_id": 30002537,
                            "type_id": 35835,
                            "unanchors_at": None,
                        },
                        {
                            "corporation_id": 2001,
                            "fuel_expires": dt.datetime(2020, 3, 5, 5, tzinfo=utc),
                            "next_reinforce_apply": None,
                            "next_reinforce_hour": None,
                            "profile_id": 101853,
                            "reinforce_hour": 18,
                            "services": [
                                {"name": "Clone Bay", "state": "online"},
                                {"name": "Market Hub", "state": "offline"},
                            ],
                            "state": "shield_vulnerable",
                            "state_timer_end": dt.datetime(2020, 4, 5, 7, tzinfo=utc),
                            "state_timer_start": dt.datetime(
                                2020, 4, 5, 6, 30, tzinfo=utc
                            ),
                            "structure_id": 1000000000001,
                            "system_id": 30002537,
                            "type_id": 35832,
                            "unanchors_at": dt.datetime(2020, 5, 5, 6, 30, tzinfo=utc),
                        },
                        {
                            "corporation_id": 2001,
                            "fuel_expires": None,
                            "next_reinforce_apply": None,
                            "next_reinforce_hour": None,
                            "profile_id": 101853,
                            "reinforce_hour": 18,
                            "services": None,
                            "state": "shield_vulnerable",
                            "state_timer_end": None,
                            "state_timer_start": None,
                            "structure_id": 1000000000003,
                            "system_id": 30000476,
                            "type_id": 35832,
                            "unanchors_at": None,
                        },
                    ],
                    "2005": [],
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_starbases",
                "corporation_id",
                needs_token=True,
                data={
                    "2001": [
                        {
                            "moon_id": 40161465,
                            "starbase_id": 1300000000001,
                            "state": "online",
                            "system_id": 30002537,
                            "type_id": 16213,
                            "reinforced_until": dt.datetime(2020, 4, 5, 7, tzinfo=utc),
                        },
                        {
                            "moon_id": 40161466,
                            "starbase_id": 1300000000002,
                            "state": "offline",
                            "system_id": 30002537,
                            "type_id": 20061,
                            "unanchors_at": dt.datetime(2020, 5, 5, 7, tzinfo=utc),
                        },
                        {
                            "moon_id": 40029527,
                            "reinforced_until": dt.datetime(2020, 1, 2, 3, tzinfo=utc),
                            "starbase_id": 1300000000003,
                            "state": "reinforced",
                            "system_id": 30000474,
                            "type_id": 20062,
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Corporation",
                "get_corporations_corporation_id_starbases_starbase_id",
                ("corporation_id", "starbase_id"),
                needs_token=True,
                data={
                    "2001": {
                        "1300000000001": {
                            "allow_alliance_members": True,
                            "allow_corporation_members": True,
                            "anchor": "config_starbase_equipment_role",
                            "attack_if_at_war": False,
                            "attack_if_other_security_status_dropping": False,
                            "fuel_bay_take": "config_starbase_equipment_role",
                            "fuel_bay_view": "starbase_fuel_technician_role",
                            "fuels": [
                                {"quantity": 960, "type_id": 4051},
                                {"quantity": 11678, "type_id": 16275},
                            ],
                            "offline": "config_starbase_equipment_role",
                            "online": "config_starbase_equipment_role",
                            "unanchor": "config_starbase_equipment_role",
                            "use_alliance_standings": True,
                        },
                        "1300000000002": {
                            "allow_alliance_members": True,
                            "allow_corporation_members": True,
                            "anchor": "config_starbase_equipment_role",
                            "attack_if_at_war": False,
                            "attack_if_other_security_status_dropping": False,
                            "fuel_bay_take": "config_starbase_equipment_role",
                            "fuels": [
                                {"quantity": 5, "type_id": 4051},
                                {"quantity": 11678, "type_id": 16275},
                            ],
                            "fuel_bay_view": "starbase_fuel_technician_role",
                            "offline": "config_starbase_equipment_role",
                            "online": "config_starbase_equipment_role",
                            "unanchor": "config_starbase_equipment_role",
                            "use_alliance_standings": True,
                        },
                        "1300000000003": {
                            "allow_alliance_members": True,
                            "allow_corporation_members": True,
                            "anchor": "config_starbase_equipment_role",
                            "attack_if_at_war": False,
                            "attack_if_other_security_status_dropping": False,
                            "fuel_bay_take": "config_starbase_equipment_role",
                            "fuel_bay_view": "starbase_fuel_technician_role",
                            "fuels": [
                                {"quantity": 1000, "type_id": 4051},
                                {"quantity": 11678, "type_id": 16275},
                            ],
                            "offline": "config_starbase_equipment_role",
                            "online": "config_starbase_equipment_role",
                            "unanchor": "config_starbase_equipment_role",
                            "use_alliance_standings": True,
                        },
                    }
                },
            ),
            EsiEndpoint(
                "Planetary_Interaction",
                "get_corporations_corporation_id_customs_offices",
                "corporation_id",
                needs_token=True,
                data={
                    "2001": [
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.3,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000003,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30002537,
                            "terrible_standing_tax_rate": 0.5,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000004,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000005,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000006,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                        {
                            "alliance_tax_rate": 0.02,
                            "allow_access_with_standings": True,
                            "allow_alliance_access": True,
                            "bad_standing_tax_rate": 0.02,
                            "corporation_tax_rate": 0.02,
                            "excellent_standing_tax_rate": 0.02,
                            "good_standing_tax_rate": 0.02,
                            "neutral_standing_tax_rate": 0.02,
                            "office_id": 1200000000099,
                            "reinforce_exit_end": 21,
                            "reinforce_exit_start": 19,
                            "standing_level": "terrible",
                            "system_id": 30000474,
                            "terrible_standing_tax_rate": 0.02,
                        },
                    ]
                },
            ),
            EsiEndpoint(
                "Universe",
                "get_universe_structures_structure_id",
                "structure_id",
                needs_token=True,
                data={
                    "1000000000001": {
                        "corporation_id": 2001,
                        "name": "Amamake - Test Structure Alpha",
                        "position": {
                            "x": 55028384780.0,
                            "y": 7310316270.0,
                            "z": -163686684205.0,
                        },
                        "solar_system_id": 30002537,
                        "type_id": 35832,
                    },
                    "1000000000002": {
                        "corporation_id": 2001,
                        "name": "Amamake - Test Structure Bravo",
                        "position": {
                            "x": -2518743930339.066,
                            "y": -130157937025.56424,
                            "z": -442026427345.6355,
                        },
                        "solar_system_id": 30002537,
                        "type_id": 35835,
                    },
                    "1000000000003": {
                        "corporation_id": 2001,
                        "name": "Amamake - Test Structure Charlie",
                        "position": {
                            "x": -2518743930339.066,
                            "y": -130157937025.56424,
                            "z": -442026427345.6355,
                        },
                        "solar_system_id": 30000476,
                        "type_id": 35832,
                    },
                },
            ),
        ]
        cls.esi_client_stub = EsiClientStub.create_from_endpoints(cls.endpoints)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_can_sync_upwell_structures(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertAlmostEqual(
            owner.structures_last_update_at, now(), delta=dt.timedelta(seconds=30)
        )

        # must contain all expected structures
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # verify attributes for structure
        structure = Structure.objects.get(id=1000000000001)
        self.assertEqual(structure.name, "Test Structure Alpha")
        self.assertEqual(structure.position_x, 55028384780.0)
        self.assertEqual(structure.position_y, 7310316270.0)
        self.assertEqual(structure.position_z, -163686684205.0)
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(structure.eve_type_id, 35832)
        self.assertEqual(int(structure.owner.corporation.corporation_id), 2001)
        self.assertEqual(structure.state, Structure.State.SHIELD_VULNERABLE)
        self.assertEqual(structure.reinforce_hour, 18)
        self.assertEqual(
            structure.fuel_expires_at, dt.datetime(2020, 3, 5, 5, 0, 0, tzinfo=utc)
        )
        self.assertEqual(
            structure.state_timer_start, dt.datetime(2020, 4, 5, 6, 30, 0, tzinfo=utc)
        )
        self.assertEqual(
            structure.state_timer_end, dt.datetime(2020, 4, 5, 7, 0, 0, tzinfo=utc)
        )
        self.assertEqual(
            structure.unanchors_at, dt.datetime(2020, 5, 5, 6, 30, 0, tzinfo=utc)
        )

        # must have created services with localizations
        # structure 1000000000001
        expected = {
            to_json(
                {
                    "name": "Clone Bay",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Clone Bay_zh",
                    "state": StructureService.State.ONLINE,
                }
            ),
            to_json(
                {
                    "name": "Market Hub",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Market Hub_zh",
                    "state": StructureService.State.OFFLINE,
                }
            ),
        }
        structure = Structure.objects.get(id=1000000000001)
        services = {
            to_json(
                {
                    "name": x.name,
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": x.name_zh,
                    "state": x.state,
                }
            )
            for x in structure.services.all()
        }
        self.assertEqual(services, expected)

        # must have created services with localizations
        # structure 1000000000002
        expected = {
            to_json(
                {
                    "name": "Reprocessing",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Reprocessing_zh",
                    "state": StructureService.State.ONLINE,
                }
            ),
            to_json(
                {
                    "name": "Moon Drilling",
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": "Moon Drilling_zh",
                    "state": StructureService.State.ONLINE,
                }
            ),
        }
        structure = Structure.objects.get(id=1000000000002)
        services = {
            to_json(
                {
                    "name": x.name,
                    "name_de": "",
                    "name_ko": "",
                    "name_ru": "",
                    # "name_zh": x.name_zh,
                    "state": x.state,
                }
            )
            for x in structure.services.all()
        }
        self.assertEqual(services, expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_can_sync_pocos(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        # must contain all expected structures
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)
        self.assertSetEqual(
            set(PocoDetails.objects.values_list("structure_id", flat=True)),
            {1200000000003, 1200000000004, 1200000000005, 1200000000006, 1200000000099},
        )

        # verify attributes for POCO
        structure = Structure.objects.get(id=1200000000003)
        self.assertEqual(structure.name, "Planet (Barren)")
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(int(structure.owner.corporation.corporation_id), 2001)
        self.assertEqual(structure.eve_type_id, 2233)
        self.assertEqual(structure.reinforce_hour, 20)
        self.assertEqual(structure.state, Structure.State.UNKNOWN)
        self.assertEqual(structure.eve_planet_id, 40161472)

        # verify attributes for POCO details
        details = structure.poco_details
        self.assertEqual(details.alliance_tax_rate, 0.02)
        self.assertTrue(details.allow_access_with_standings)
        self.assertTrue(details.allow_alliance_access)
        self.assertEqual(details.bad_standing_tax_rate, 0.3)
        self.assertEqual(details.corporation_tax_rate, 0.02)
        self.assertEqual(details.excellent_standing_tax_rate, 0.02)
        self.assertEqual(details.good_standing_tax_rate, 0.02)
        self.assertEqual(details.neutral_standing_tax_rate, 0.02)
        self.assertEqual(details.reinforce_exit_end, 21)
        self.assertEqual(details.reinforce_exit_start, 19)
        self.assertEqual(details.standing_level, PocoDetails.StandingLevel.TERRIBLE)
        self.assertEqual(details.terrible_standing_tax_rate, 0.5)

        # empty name for POCO with no asset data
        structure = Structure.objects.get(id=1200000000099)
        self.assertEqual(structure.name, "")

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_can_sync_starbases(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        # must contain all expected structures
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1300000000001,
            1300000000002,
            1300000000003,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

        # verify attributes for POS
        structure = Structure.objects.get(id=1300000000001)
        self.assertEqual(structure.name, "Home Sweat Home")
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(int(structure.owner.corporation.corporation_id), 2001)
        self.assertEqual(structure.eve_type_id, 16213)
        self.assertEqual(structure.state, Structure.State.POS_ONLINE)
        self.assertEqual(structure.eve_moon_id, 40161465)
        self.assertEqual(
            structure.state_timer_end, dt.datetime(2020, 4, 5, 7, 0, 0, tzinfo=utc)
        )
        self.assertAlmostEqual(
            structure.fuel_expires_at,
            now() + dt.timedelta(hours=24),
            delta=dt.timedelta(seconds=30),
        )
        self.assertEqual(structure.position_x, 40.2)
        self.assertEqual(structure.position_y, 27.3)
        self.assertEqual(structure.position_z, -19.4)
        # verify details
        detail = structure.starbase_detail
        self.assertTrue(detail.allow_alliance_members)
        self.assertTrue(detail.allow_corporation_members)
        self.assertEqual(
            detail.anchor_role, StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
        )
        self.assertFalse(detail.attack_if_at_war)
        self.assertFalse(detail.attack_if_other_security_status_dropping)
        self.assertEqual(
            detail.anchor_role, StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE
        )
        self.assertEqual(
            detail.fuel_bay_take_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.fuel_bay_view_role,
            StarbaseDetail.Role.STARBASE_FUEL_TECHNICIAN_ROLE,
        )
        self.assertEqual(
            detail.offline_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.online_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertEqual(
            detail.unanchor_role,
            StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        )
        self.assertTrue(detail.use_alliance_standings)
        # fuels
        self.assertEqual(detail.fuels.count(), 2)
        self.assertEqual(detail.fuels.get(eve_type_id=4051).quantity, 960)
        self.assertEqual(detail.fuels.get(eve_type_id=16275).quantity, 11678)

        structure = Structure.objects.get(id=1300000000002)
        self.assertEqual(structure.name, "Bat cave")
        self.assertEqual(structure.eve_solar_system_id, 30002537)
        self.assertEqual(int(structure.owner.corporation.corporation_id), 2001)
        self.assertEqual(structure.eve_type_id, 20061)
        self.assertEqual(structure.state, Structure.State.POS_OFFLINE)
        self.assertEqual(structure.eve_moon_id, 40161466)
        self.assertEqual(
            structure.unanchors_at, dt.datetime(2020, 5, 5, 7, 0, 0, tzinfo=utc)
        )
        self.assertIsNone(structure.fuel_expires_at)
        self.assertFalse(structure.generatednotification_set.exists())

        structure = Structure.objects.get(id=1300000000003)
        self.assertEqual(structure.name, "Panic Room")
        self.assertEqual(structure.eve_solar_system_id, 30000474)
        self.assertEqual(int(structure.owner.corporation.corporation_id), 2001)
        self.assertEqual(structure.eve_type_id, 20062)
        self.assertEqual(structure.state, Structure.State.POS_REINFORCED)
        self.assertEqual(structure.eve_moon_id, 40029527)
        self.assertAlmostEqual(
            structure.fuel_expires_at,
            now() + dt.timedelta(hours=133, minutes=20),
            delta=dt.timedelta(seconds=30),
        )
        self.assertEqual(
            structure.state_timer_end, dt.datetime(2020, 1, 2, 3, tzinfo=utc)
        )
        self.assertTrue(structure.generatednotification_set.exists())

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    @patch(MODULE_PATH + ".notify", spec=True)
    def test_can_sync_all_structures_and_notify_user(self, mock_notify, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)

        # when
        owner.update_structures_esi(user=self.user)

        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)

        # must contain all expected structures
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
            1300000000001,
            1300000000002,
            1300000000003,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

        # user report has been sent
        self.assertTrue(mock_notify.called)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_can_handle_owner_without_structures(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        user, _ = create_user_from_evecharacter(
            1005,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner = create_owner_from_user(user)  # corp_ID = 2005
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_structure_sync_fresh)
        self.assertSetEqual(owner.structures.ids(), set())

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_should_not_break_when_endpoint_for_fetching_upwell_structures_is_down(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_structures",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_should_not_break_on_http_error_when_fetching_custom_offices(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Planetary_Interaction",
            "get_corporations_corporation_id_customs_offices",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_should_not_break_on_http_error_when_fetching_custom_office_names(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Assets",
            "post_corporations_corporation_id_assets_names",
            http_error_code=404,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_not_break_on_http_error_when_fetching_star_bases(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".notify", spec=True)
    def test_should_mark_error_when_character_not_director_while_updating_starbases(
        self, mock_notify, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=403,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        self.assertTrue(mock_notify)
        character = owner.characters.first()
        self.assertEqual(character.error_count, 1)

    @patch(MODULE_PATH + ".STRUCTURES_ESI_DIRECTOR_ERROR_MAX_RETRIES", 3)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".notify", spec=True)
    def test_should_remove_character_when_not_director_while_updating_starbases(
        self, mock_notify, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=403,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        character = owner.characters.first()
        character.error_count = 3
        character.save()
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        self.assertFalse(owner.is_structure_sync_fresh)
        self.assertTrue(mock_notify)
        self.assertEqual(owner.characters.count(), 0)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_update_will_not_break_on_http_error_from_structure_info(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Universe", "get_universe_structures_structure_id", http_error_code=500
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=1000000000002)
        self.assertEqual(structure.name, "(no data)")

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".Structure.objects.update_or_create_from_dict")
    def test_update_will_not_break_on_http_error_when_creating_structures(
        self, mock_create_structure, mock_esi
    ):
        mock_create_structure.side_effect = OSError
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_remove_old_upwell_structures(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        create_upwell_structure(owner=owner, id=1000000000004, name="delete-me")
        # when
        owner.update_structures_esi()
        # then
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_should_remove_old_pocos(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        create_poco(owner=owner, id=1000000000004, name="delete-me")
        # when
        owner.update_structures_esi()
        # then
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1200000000003,
            1200000000004,
            1200000000005,
            1200000000006,
            1200000000099,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_remove_old_starbases(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        create_starbase(owner=owner, id=1300000000099, name="delete-me")
        # when
        owner.update_structures_esi()
        # then
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1300000000001,
            1300000000002,
            1300000000003,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_tags_are_not_modified_by_update(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then

        # should contain the right structures
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # adding tags
        tag_a = StructureTag.objects.get(name="tag_a")
        s = Structure.objects.get(id=1000000000001)
        s.tags.add(tag_a)
        s.save()

        # run update task 2nd time
        owner.update_structures_esi()

        # should still contain alls structures
        expected = {1000000000001, 1000000000002, 1000000000003}
        self.assertSetEqual(owner.structures.ids(), expected)

        # should still contain the tag
        s_new = Structure.objects.get(id=1000000000001)
        self.assertEqual(s_new.tags.get(name="tag_a"), tag_a)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_not_delete_existing_upwell_structures_when_update_failed(
        self, mock_esi
    ):
        # given
        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_structures",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        create_upwell_structure(owner=owner, id=1000000000001)
        create_upwell_structure(owner=owner, id=1000000000002)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {1000000000001, 1000000000002}
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_should_not_delete_existing_pocos_when_update_failed(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Planetary_Interaction",
            "get_corporations_corporation_id_customs_offices",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        create_poco(owner=owner, id=1200000000003)
        create_poco(owner=owner, id=1200000000004)
        # when
        owner.update_structures_esi()
        # then
        self.assertFalse(owner.is_structure_sync_fresh)
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1200000000003,
            1200000000004,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_not_delete_existing_starbases_when_update_failed(self, mock_esi):
        # given

        new_endpoint = EsiEndpoint(
            "Corporation",
            "get_corporations_corporation_id_starbases",
            http_error_code=500,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        create_starbase(owner=owner, id=1300000000001)
        create_starbase(owner=owner, id=1300000000002)
        # when
        owner.update_structures_esi()
        # then
        # self.assertFalse(owner.is_structure_sync_fresh)
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1300000000001,
            1300000000002,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_remove_outdated_services(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        structure = create_upwell_structure(owner=owner, id=1000000000002)
        create_structure_service(structure=structure, name="Clone Bay")
        # when
        owner.update_structures_esi()
        # then
        structure.refresh_from_db()
        services = {
            obj.name for obj in StructureService.objects.filter(structure=structure)
        }
        self.assertEqual(services, {"Moon Drilling", "Reprocessing"})

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_should_have_empty_name_if_not_match_with_planets(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        EvePlanet.objects.all().delete()
        # when
        owner.update_structures_esi()
        # then
        self.assertTrue(owner.is_structure_sync_fresh)
        structure = Structure.objects.get(id=1200000000003)
        self.assertEqual(structure.name, "")

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    def test_define_poco_name_from_planet_type_if_found(self, mock_esi):
        # given
        mock_esi.client = self.esi_client_stub
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        structure = Structure.objects.get(id=1200000000003)
        self.assertEqual(structure.eve_planet_id, 40161472)
        self.assertEqual(structure.name, "Planet (Barren)")

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(
        "structures.models.structures_1.STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
    )
    @patch("structures.models.notifications.Webhook.send_message")
    def test_should_send_refueled_notification_when_fuel_level_increased(
        self, mock_send_message, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_send_message.return_value = 1
        webhook = Webhook.objects.create(
            name="Webhook 1",
            url="webhook-1",
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
            is_active=True,
        )
        owner = create_owner_from_user(self.user)
        owner.webhooks.add(webhook)
        owner.update_structures_esi()
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = dt.datetime(2020, 3, 3, 0, 0, tzinfo=utc)
        structure.save()
        # when
        with patch("structures.models.structures_1.now") as now:
            now.return_value = dt.datetime(2020, 3, 2, 0, 0, tzinfo=utc)
            owner.update_structures_esi()
        # then
        self.assertTrue(mock_send_message.called)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(
        "structures.models.structures_1.STRUCTURES_FEATURE_REFUELED_NOTIFICATIONS", True
    )
    @patch("structures.models.notifications.Webhook.send_message")
    def test_should_not_send_refueled_notification_when_fuel_level_unchanged(
        self, mock_send_message, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_send_message.side_effect = RuntimeError
        webhook = Webhook.objects.create(
            name="Webhook 1",
            url="webhook-1",
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
            is_active=True,
        )
        owner = create_owner_from_user(self.user)
        owner.webhooks.add(webhook)
        with patch("structures.models.structures_1.now") as now:
            now.return_value = dt.datetime(2020, 3, 2, 0, 0, tzinfo=utc)
            owner.update_structures_esi()
            # when
            owner.update_structures_esi()
        # then
        self.assertFalse(mock_send_message.called)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch("structures.models.notifications.Webhook.send_message")
    def test_should_remove_outdated_fuel_alerts_when_fuel_level_changed(
        self, mock_send_message, mock_esi
    ):
        # given
        mock_esi.client = self.esi_client_stub
        mock_send_message.return_value = 1
        webhook = Webhook.objects.create(
            name="Webhook 1",
            url="webhook-1",
            notification_types=[NotificationType.STRUCTURE_REFUELED_EXTRA],
            is_active=True,
        )
        owner = create_owner_from_user(self.user)
        owner.webhooks.add(webhook)
        owner.update_structures_esi()
        structure = Structure.objects.get(id=1000000000001)
        structure.fuel_expires_at = dt.datetime(2020, 3, 3, 0, 0, tzinfo=utc)
        structure.save()
        config = FuelAlertConfig.objects.create(start=48, end=0, repeat=12)
        structure.structure_fuel_alerts.create(config=config, hours=12)
        # when
        with patch("structures.models.structures_1.now") as now:
            now.return_value = dt.datetime(2020, 3, 2, 0, 0, tzinfo=utc)
            owner.update_structures_esi()
        # then
        self.assertEqual(structure.structure_fuel_alerts.count(), 0)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    def test_should_not_break_when_starbase_names_not_found(self, mock_esi):
        # given
        new_endpoint = EsiEndpoint(
            "Assets",
            "post_corporations_corporation_id_assets_names",
            http_error_code=404,
        )
        mock_esi.client = self.esi_client_stub.replace_endpoints([new_endpoint])
        owner = create_owner_from_user(self.user)
        # when
        owner.update_structures_esi()
        # then
        owner.refresh_from_db()
        expected = {
            1000000000001,
            1000000000002,
            1000000000003,
            1300000000001,
            1300000000002,
            1300000000003,
        }
        self.assertSetEqual(owner.structures.ids(), expected)

    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    # @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    # def test_should_notify_admins_when_service_is_restored(
    #     self, mock_esi_client
    # ):
    #     # given
    #     mock_esi_client.side_effect = esi_mock_client
    #     owner = create_owner_from_user(self.user)
    #     owner.is_structure_sync_fresh = False
    #     owner.save()
    #     # when
    #     owner.update_structures_esi()
    #     # then
    #     owner.refresh_from_db()
    #     self.assertTrue(owner.is_structure_sync_fresh)



from typing import List

from django.test import RequestFactory

from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from structures.core.serializers import PocoListSerializer, StructureListSerializer
from structures.models import Structure
from structures.tests.testdata.factories import (
    create_owner_from_user,
    create_poco,
    create_starbase,
    create_upwell_structure,
)
from structures.tests.testdata.helpers import load_entities
from structures.tests.testdata.load_eveuniverse import load_eveuniverse


def to_dict(lst: List[dict], key="id"):
    return {obj[key]: obj for obj in lst}


class TestStructureListSerializer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        load_eveuniverse()
        cls.user, _ = create_user_from_evecharacter(1001)
        cls.owner = create_owner_from_user(cls.user)
        cls.request = cls.factory.get("/")
        cls.request.user = cls.user

    def test_should_show_not_reinforced_for_structure(self):
        # given
        structure = create_upwell_structure(owner=self.owner)
        # when
        data = StructureListSerializer(
            queryset=Structure.objects.all(), request=self.request
        ).to_list()
        # then
        obj = to_dict(data)[structure.id]
        self.assertFalse(obj["is_reinforced"])

    def test_should_show_reinforced_for_structure(self):
        # given
        structure = create_upwell_structure(
            owner=self.owner, state=Structure.State.ARMOR_REINFORCE
        )
        # when
        data = StructureListSerializer(
            queryset=Structure.objects.all(), request=self.request
        ).to_list()
        # then
        obj = to_dict(data)[structure.id]
        self.assertTrue(obj["is_reinforced"])

    def test_should_show_not_reinforced_for_starbase(self):
        # given
        structure = create_starbase(owner=self.owner)
        # when
        data = StructureListSerializer(
            queryset=Structure.objects.all(), request=self.request
        ).to_list()
        # then
        obj = to_dict(data)[structure.id]
        self.assertFalse(obj["is_reinforced"])

    def test_should_show_reinforced_for_starbase(self):
        # given
        structure = create_starbase(
            owner=self.owner, state=Structure.State.POS_REINFORCED
        )
        # when
        data = StructureListSerializer(
            queryset=Structure.objects.all(), request=self.request
        ).to_list()
        # then
        obj = to_dict(data)[structure.id]
        self.assertTrue(obj["is_reinforced"])


class TestPocoListSerializer(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.factory = RequestFactory()
        load_entities()
        load_eveuniverse()
        cls.user, _ = create_user_from_evecharacter(1001)
        cls.owner = create_owner_from_user(cls.user)
        cls.request = cls.factory.get("/")
        cls.request.user = cls.user

    def test_should_extract_planet_type(self):
        # given
        structure = create_poco(owner=self.owner)
        # when
        data = PocoListSerializer(
            queryset=Structure.objects.all(), request=self.request
        ).to_list()
        # then
        obj = to_dict(data)[structure.id]
        self.assertEqual(obj["planet_type_name"], "Barren")



import datetime as dt
from datetime import datetime, timedelta
from unittest.mock import patch

from django.utils.timezone import now, utc
from esi.errors import TokenError
from esi.models import Token
from eveuniverse.models import EveSolarSystem

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter

from structures.models import Owner, OwnerCharacter
from structures.tests.testdata.factories import create_owner_from_user
from structures.tests.testdata.helpers import (
    create_structures,
    load_entities,
    set_owner_character,
)
from structures.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "structures.models.owners"


class TestOwner(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)

    def setUp(self) -> None:
        self.owner.is_alliance_main = True
        self.owner.save()

    def test_str(self):
        # when
        result = str(self.owner)
        # then
        self.assertEqual(result, "Wayne Technologies")

    def test_repr(self):
        # when
        result = repr(self.owner)
        # then
        self.assertEqual(
            result, f"Owner(pk={self.owner.pk}, corporation='Wayne Technologies')"
        )

    @patch(MODULE_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    def test_is_structure_sync_fresh(self):
        # no errors and recent sync
        self.owner.structures_last_update_at = now()
        self.assertTrue(self.owner.is_structure_sync_fresh)

        # no errors and sync within grace period
        self.owner.structures_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_structure_sync_fresh)

        # no error, but no sync within grace period
        self.owner.structures_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_structure_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_is_notification_sync_fresh(self):
        # no errors and recent sync
        self.owner.notifications_last_update_at = now()
        self.assertTrue(self.owner.is_notification_sync_fresh)

        # no errors and sync within grace period
        self.owner.notifications_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_notification_sync_fresh)

        # no error, but no sync within grace period
        self.owner.notifications_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_notification_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_is_forwarding_sync_fresh(self):
        # no errors and recent sync
        self.owner.forwarding_last_update_at = now()
        self.assertTrue(self.owner.is_forwarding_sync_fresh)

        # no errors and sync within grace period
        self.owner.forwarding_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_forwarding_sync_fresh)

        # no error, but no sync within grace period
        self.owner.forwarding_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_forwarding_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    def test_is_assets_sync_fresh(self):
        # no errors and recent sync
        self.owner.assets_last_update_at = now()
        self.assertTrue(self.owner.is_assets_sync_fresh)

        # no errors and sync within grace period
        self.owner.assets_last_update_at = now() - timedelta(minutes=29)
        self.assertTrue(self.owner.is_assets_sync_fresh)

        # no error, but no sync within grace period
        self.owner.assets_last_update_at = now() - timedelta(minutes=31)
        self.assertFalse(self.owner.is_assets_sync_fresh)

    @patch(MODULE_PATH + ".STRUCTURES_STRUCTURE_SYNC_GRACE_MINUTES", 30)
    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    @patch(MODULE_PATH + ".STRUCTURES_NOTIFICATION_SYNC_GRACE_MINUTES", 30)
    def test_is_all_syncs_ok(self):
        self.owner.structures_last_update_at = now()
        self.owner.notifications_last_update_at = now()
        self.owner.forwarding_last_update_at = now()
        self.owner.assets_last_update_at = now()
        self.assertTrue(self.owner.are_all_syncs_ok)

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    def test_get_esi_scopes_pocos_off(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", False)
    def test_get_esi_scopes_pocos_on(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-planets.read_customs_offices.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", False)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    def test_get_esi_scopes_starbases_on(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_starbases.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_CUSTOMS_OFFICES", True)
    @patch(MODULE_PATH + ".STRUCTURES_FEATURE_STARBASES", True)
    def test_get_esi_scopes_starbases_and_custom_offices(self):
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            {
                "esi-corporations.read_structures.v1",
                "esi-universe.read_structures.v1",
                "esi-characters.read_notifications.v1",
                "esi-corporations.read_starbases.v1",
                "esi-planets.read_customs_offices.v1",
                "esi-assets.read_corporation_assets.v1",
            },
        )

    def test_should_ensure_only_one_owner_is_alliance_main_1(self):
        # given
        self.assertTrue(self.owner.is_alliance_main)
        owner = Owner.objects.get(corporation__corporation_id=2002)
        # when
        owner.is_alliance_main = True
        owner.save()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_alliance_main)
        self.owner.refresh_from_db()
        self.assertFalse(self.owner.is_alliance_main)

    def test_should_ensure_only_one_owner_is_alliance_main_2(self):
        # given
        self.assertTrue(self.owner.is_alliance_main)
        owner = Owner.objects.get(corporation__corporation_id=2007)
        # when
        owner.is_alliance_main = True
        owner.save()
        # then
        owner.refresh_from_db()
        self.assertTrue(owner.is_alliance_main)
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_alliance_main)

    def test_should_ensure_only_one_owner_is_alliance_main_3(self):
        # given
        self.assertTrue(self.owner.is_alliance_main)
        owner_2103 = Owner.objects.get(corporation__corporation_id=2103)
        owner_2103.is_alliance_main = True
        owner_2103.save()
        owner_2102 = Owner.objects.get(corporation__corporation_id=2102)
        # when
        owner_2102.is_alliance_main = True
        owner_2102.save()
        # then
        owner_2102.refresh_from_db()
        self.assertTrue(owner_2102.is_alliance_main)
        self.owner.refresh_from_db()
        self.assertTrue(self.owner.is_alliance_main)
        owner_2103.refresh_from_db()
        self.assertTrue(owner_2103.is_alliance_main)


class TestOwnerHasSov(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        load_entities()
        user, _ = create_user_from_evecharacter(
            1001, permissions=["structures.add_structure_owner"]
        )
        cls.owner = create_owner_from_user(user=user)

    def test_should_return_true_when_owner_has_sov(self):
        # given
        system = EveSolarSystem.objects.get(name="1-PGSG")
        # when/then
        self.assertTrue(self.owner.has_sov(system))

    def test_should_return_false_when_owner_has_no_sov(self):
        # given
        system = EveSolarSystem.objects.get(name="A-C5TC")
        # when/then
        self.assertFalse(self.owner.has_sov(system))

    def test_should_return_false_when_owner_is_outside_nullsec(self):
        # given
        system = EveSolarSystem.objects.get(name="Amamake")
        # when/then
        self.assertFalse(self.owner.has_sov(system))


@patch(MODULE_PATH + ".notify")
@patch(MODULE_PATH + ".notify_admins")
class TestOwnerFetchToken(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities([EveCorporationInfo, EveCharacter])
        load_eveuniverse()
        cls.character = EveCharacter.objects.get(character_id=1001)
        cls.corporation = EveCorporationInfo.objects.get(corporation_id=2001)

    def test_should_return_correct_token(self, mock_notify_admins, mock_notify):
        # given
        user, character_ownership = create_user_from_evecharacter(
            1001,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner = Owner.objects.create(corporation=self.corporation)
        owner.add_character(character_ownership)
        # when
        token = owner.fetch_token()
        # then
        self.assertIsInstance(token, Token)
        self.assertEqual(token.user, user)
        self.assertEqual(token.character_id, 1001)
        self.assertSetEqual(
            set(Owner.get_esi_scopes()),
            set(token.scopes.values_list("name", flat=True)),
        )
        self.assertFalse(mock_notify_admins.called)
        self.assertFalse(mock_notify.called)
        self.assertEqual(owner.characters.count(), 1)

    def test_raise_error_when_no_sync_char_defined(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = Owner.objects.create(corporation=self.corporation)
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertFalse(mock_notify_admins.called)
        self.assertFalse(mock_notify.called)

    def test_raise_error_when_user_has_no_permission_and_delete_character(
        self, mock_notify_admins, mock_notify
    ):
        # given
        _, character_ownership = create_user_from_evecharacter(
            1001, scopes=Owner.get_esi_scopes()
        )
        owner = Owner.objects.create(corporation=self.corporation)
        owner.add_character(character_ownership)
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 0)

    def test_raise_error_when_token_not_found_and_delete_character(
        self, mock_notify_admins, mock_notify
    ):
        # given
        with patch(MODULE_PATH + ".Token.objects.filter") as my_mock:
            my_mock.return_value = None
            user, character_ownership = create_user_from_evecharacter(
                1001,
                permissions=["structures.add_structure_owner"],
                scopes=Owner.get_esi_scopes(),
            )
        owner = Owner.objects.create(corporation=self.corporation)
        owner.add_character(character_ownership)
        user.token_set.first().scopes.clear()
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 0)

    def test_raise_error_when_character_no_longer_a_corporation_member_and_delete_it(
        self, mock_notify_admins, mock_notify
    ):
        # given
        _, character_ownership = create_user_from_evecharacter(
            1011,
            scopes=Owner.get_esi_scopes(),
            permissions=["structures.add_structure_owner"],
        )
        owner = Owner.objects.create(corporation=self.corporation)
        owner.characters.create(character_ownership=character_ownership)
        # when/then
        with self.assertRaises(TokenError):
            owner.fetch_token()
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 0)

    def test_should_delete_invalid_characters_and_return_token_from_valid_char(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2102)
        )
        user, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1011)
        my_character.notifications_last_used_at = dt.datetime(
            2021, 1, 1, 1, 2, tzinfo=utc
        )
        my_character.save()
        _, character_ownership_1102 = create_user_from_evecharacter(
            1102, scopes=Owner.get_esi_scopes()
        )
        my_character = owner.add_character(character_ownership_1102)
        my_character.notifications_last_used_at = dt.datetime(
            2021, 1, 1, 1, 1, tzinfo=utc
        )
        my_character.save()
        # when
        token = owner.fetch_token()
        # then
        self.assertIsInstance(token, Token)
        self.assertEqual(token.user, user)
        self.assertTrue(mock_notify_admins.called)
        self.assertTrue(mock_notify.called)
        self.assertEqual(owner.characters.count(), 1)

    def test_should_rotate_through_characters_for_notification(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2102)
        )
        _, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1011)
        my_character.notifications_last_used_at = dt.datetime(
            2021, 1, 1, 0, 5, tzinfo=utc
        )
        my_character.save()
        _, character_ownership_1102 = create_user_from_evecharacter(
            1102,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1102)
        my_character.notifications_last_used_at = dt.datetime(
            2021, 1, 1, 0, 0, tzinfo=utc
        )
        my_character.save()
        tokens_received = list()
        # when
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS,
                ignore_schedule=True,
            ).character_id
        )
        # then
        self.assertListEqual(tokens_received, [1102, 1011, 1102, 1011])

    def test_should_rotate_through_characters_for_structures(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2102)
        )
        _, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1011)
        my_character.structures_last_used_at = dt.datetime(2021, 1, 1, 0, 5, tzinfo=utc)
        my_character.save()
        _, character_ownership_1102 = create_user_from_evecharacter(
            1102,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1102)
        my_character.structures_last_used_at = dt.datetime(2021, 1, 1, 0, 0, tzinfo=utc)
        my_character.save()
        tokens_received = list()
        # when
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.STRUCTURES,
                ignore_schedule=True,
            ).character_id
        )
        # then
        self.assertListEqual(tokens_received, [1102, 1011, 1102, 1011])

    def test_should_rotate_through_characters_based_on_schedule(
        self, mock_notify_admins, mock_notify
    ):
        # given
        owner = Owner.objects.create(
            corporation=EveCorporationInfo.objects.get(corporation_id=2102)
        )
        _, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1011)
        my_character.notifications_last_used_at = dt.datetime(
            2021, 1, 1, 0, 1, tzinfo=utc
        )
        my_character.save()
        _, character_ownership_1102 = create_user_from_evecharacter(
            1102,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        my_character = owner.add_character(character_ownership_1102)
        my_character.notifications_last_used_at = dt.datetime(
            2021, 1, 1, 0, 0, tzinfo=utc
        )
        my_character.save()
        tokens_received = list()
        # when
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS
            ).character_id
        )
        tokens_received.append(
            owner.fetch_token(
                rotate_characters=Owner.RotateCharactersType.NOTIFICATIONS
            ).character_id
        )
        # then
        self.assertListEqual(tokens_received, [1102, 1011, 1102, 1102])


class TestOwnerCharacters(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()
        create_structures()
        cls.user, cls.owner = set_owner_character(character_id=1001)

    def test_should_add_new_character(self):
        # given
        owner = Owner.objects.get(corporation__corporation_id=2002)
        _, character_ownership = create_user_from_evecharacter(
            1003,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        # when
        result = owner.add_character(character_ownership)
        # then
        self.assertIsInstance(result, OwnerCharacter)
        self.assertEqual(result.owner, owner)
        self.assertEqual(result.character_ownership, character_ownership)
        self.assertIsNone(result.notifications_last_used_at)

    def test_should_not_overwrite_existing_characters(self):
        # given
        character = self.owner.characters.first()
        my_dt = datetime(year=2021, month=2, day=11, hour=12, tzinfo=utc)
        character.notifications_last_used_at = my_dt
        character.save()
        # when
        result = self.owner.add_character(character.character_ownership)
        # then
        self.assertIsInstance(result, OwnerCharacter)
        self.assertEqual(result.owner, self.owner)
        self.assertEqual(result.character_ownership, character.character_ownership)
        self.assertEqual(result.notifications_last_used_at, my_dt)

    def test_should_prevent_adding_character_from_other_corporation(self):
        # given
        _, character_ownership = create_user_from_evecharacter(
            1003,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        # when
        with self.assertRaises(ValueError):
            self.owner.add_character(character_ownership)

    def test_should_add_character_to_existing_set(self):
        # given
        owner = Owner.objects.get(corporation__corporation_id=2102)
        _, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner.add_character(character_ownership_1011)
        _, character_ownership_1102 = create_user_from_evecharacter(
            1102,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        # when
        owner.add_character(character_ownership_1102)
        # then
        owner_character_pks = set(
            owner.characters.values_list("character_ownership__pk", flat=True)
        )
        expected_pks = {character_ownership_1011.pk, character_ownership_1102.pk}
        self.assertSetEqual(owner_character_pks, expected_pks)

    def test_should_count_characters(self):
        # given
        owner = Owner.objects.get(corporation__corporation_id=2102)
        _, character_ownership_1011 = create_user_from_evecharacter(
            1011,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner.add_character(character_ownership_1011)
        _, character_ownership_1102 = create_user_from_evecharacter(
            1102,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner.add_character(character_ownership_1102)
        # when
        result = owner.characters_count()
        # then
        self.assertEqual(result, 2)

    def test_should_count_characters_when_empty(self):
        # given
        owner = Owner.objects.get(corporation__corporation_id=2102)
        # when
        result = owner.characters_count()
        # then
        self.assertEqual(result, 0)


@patch(MODULE_PATH + ".notify", spec=True)
@patch(MODULE_PATH + ".notify_admins", spec=True)
class TestOwnerDeleteCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_entities([EveCorporationInfo, EveCharacter])
        load_eveuniverse()
        cls.character = EveCharacter.objects.get(character_id=1001)
        cls.corporation = EveCorporationInfo.objects.get(corporation_id=2001)

    def test_should_delete_character_and_notify(self, mock_notify_admins, mock_notify):
        # given
        user, character_ownership = create_user_from_evecharacter(
            1001,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner = Owner.objects.create(corporation=self.corporation)
        character = owner.add_character(character_ownership)
        # when
        owner.delete_character(character=character, error="dummy error")
        # then
        self.assertEqual(owner.characters.count(), 0)
        self.assertTrue(mock_notify_admins.called)
        _, kwargs = mock_notify_admins.call_args
        self.assertIn("dummy error", kwargs["message"])
        self.assertEqual(kwargs["level"], "danger")
        self.assertTrue(mock_notify.called)
        _, kwargs = mock_notify.call_args
        self.assertIn("dummy error", kwargs["message"])
        self.assertEqual(kwargs["user"], user)
        self.assertEqual(kwargs["level"], "warning")

    def test_should_not_delete_when_errors_are_allowed(
        self, mock_notify_admins, mock_notify
    ):
        # given
        _, character_ownership = create_user_from_evecharacter(
            1001,
            permissions=["structures.add_structure_owner"],
            scopes=Owner.get_esi_scopes(),
        )
        owner = Owner.objects.create(corporation=self.corporation)
        character = owner.add_character(character_ownership)
        # when
        owner.delete_character(
            character=character, error="dummy error", max_allowed_errors=1
        )
        # then
        character.refresh_from_db()
        self.assertEqual(character.error_count, 1)
        self.assertFalse(mock_notify_admins.called)
        self.assertFalse(mock_notify.called)



from unittest.mock import patch

from app_utils.testing import NoSocketsTestCase

from structures.core.notification_types import NotificationType
from structures.models import GeneratedNotification, Notification

from .testdata.factories_2 import (
    GeneratedNotificationFactory,
    NotificationFactory,
    OwnerFactory,
)
from .testdata.load_eveuniverse import load_eveuniverse

MANAGERS_PATH = "structures.managers"


@patch(
    "structures.models.notifications.NotificationBase.add_or_remove_timer",
    spec=True,
)
class TestNotificationBaseAddOrRemoveTimers(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eveuniverse()

    def test_should_create_new_timers_from_notifications(
        self, mock_add_or_remove_timer_from_notification
    ):
        # given
        owner = OwnerFactory()
        NotificationFactory(
            owner=owner, notif_type=NotificationType.STRUCTURE_LOST_SHIELD
        )
        NotificationFactory(
            owner=owner, notif_type=NotificationType.WAR_CORPORATION_BECAME_ELIGIBLE
        )
        # when
        Notification.objects.add_or_remove_timers()
        # then
        self.assertEqual(mock_add_or_remove_timer_from_notification.call_count, 1)

    def test_should_create_new_timers_from_generated_notifications(
        self, mock_add_or_remove_timer_from_notification
    ):
        # given
        GeneratedNotificationFactory()
        # when
        GeneratedNotification.objects.add_or_remove_timers()
        # then
        self.assertEqual(mock_add_or_remove_timer_from_notification.call_count, 1)



import datetime as dt
from random import randint
from typing import List

import yaml

from django.contrib.auth.models import User
from django.utils.text import slugify
from django.utils.timezone import now

from allianceauth.eveonline.models import EveCorporationInfo

from structures.constants import EveTypeId
from structures.models import (
    EveSovereigntyMap,
    Notification,
    Owner,
    OwnerCharacter,
    PocoDetails,
    StarbaseDetail,
    StarbaseDetailFuel,
    Structure,
    StructureItem,
    StructureService,
    Webhook,
)


def create_eve_sovereignty_map(**kwargs):
    if "alliance_id" not in kwargs:
        kwargs["alliance_id"] = 3001
    return EveSovereigntyMap.objects.create(**kwargs)


def create_notification(**kwargs):
    """Args(optional):
    - data: details of notification as Python dict
    """
    params = {
        "notification_id": _generate_unique_id(Notification, "notification_id"),
        "sender_id": 2901,
        "timestamp": now(),
        "notif_type": "CorpBecameWarEligible",
        "last_updated": now(),
        "text": "{}",
    }
    if "owner" not in kwargs and "owner_id" not in kwargs:
        params["owner_id"] = 2001
    if "data" in kwargs:
        data = kwargs.pop("data")
        params["text"] = yaml.dump(data)
    params.update(kwargs)
    return Notification.objects.create(**params)


def create_owner_from_user(
    user: User, webhooks: List[Webhook] = None, **kwargs
) -> Owner:
    main_character = user.profile.main_character
    kwargs["corporation"] = EveCorporationInfo.objects.get(
        corporation_id=main_character.corporation_id
    )
    if "is_alliance_main" not in kwargs:
        kwargs["is_alliance_main"] = True
    owner = Owner.objects.create(**kwargs)
    if not webhooks:
        owner.webhooks.add(create_webhook())
    else:
        owner.webhooks.add(*webhooks)
    owner.characters.add(
        create_owner_character(
            owner=owner, character_ownership=main_character.character_ownership
        )
    )
    return owner


def create_owner_character(**kwargs) -> OwnerCharacter:
    return OwnerCharacter.objects.create(**kwargs)


def _create_structure(**kwargs) -> Structure:
    structure_id = _generate_unique_id(Structure, "id")
    params = {"id": structure_id, "name": f"Generated Structure #{structure_id}"}
    if "owner" not in kwargs and "owner_id" not in kwargs:
        params["owner_id"] = 2001  # Wayne Technologies
    if "eve_solar_system_id" not in kwargs and "eve_solar_system" not in kwargs:
        params["eve_solar_system_id"] = 30002537  # Amamake
    if "eve_type_id" not in kwargs and "eve_type" not in kwargs:
        params["eve_type_id"] = 35832  # Astrahus
    params.update(kwargs)
    return Structure.objects.create(**params)


def create_structure_item(**kwargs):
    params = {
        "id": _generate_unique_id(StructureItem, "id"),
        "location_flag": StructureItem.LocationFlag.CARGO,
        "is_singleton": False,
        "quantity": 1,
    }
    if "eve_type_id" not in kwargs and "eve_type" not in kwargs:
        params["eve_type_id"] = EveTypeId.LIQUID_OZONE
    params.update(kwargs)
    return StructureItem.objects.create(**params)


def create_poco(poco_details=None, **kwargs) -> Structure:
    params = {"state": Structure.State.UNKNOWN, "eve_type_id": EveTypeId.CUSTOMS_OFFICE}
    if "eve_type" in kwargs:
        del kwargs["eve_type"]
    if "eve_planet_id" not in kwargs and "eve_planet" not in kwargs:
        kwargs["eve_planet_id"] = 40161472  # Amamake V
    params.update(kwargs)
    structure = _create_structure(**params)
    assert structure.is_poco
    poco_params = {
        "alliance_tax_rate": 0.05,
        "allow_access_with_standings": True,
        "allow_alliance_access": True,
        "bad_standing_tax_rate": 0.1,
        "corporation_tax_rate": 0.1,
        "excellent_standing_tax_rate": 0.05,
        "good_standing_tax_rate": 0.05,
        "neutral_standing_tax_rate": 0.1,
        "reinforce_exit_end": 20,
        "reinforce_exit_start": 18,
        "standing_level": PocoDetails.StandingLevel.TERRIBLE,
        "structure": structure,
        "terrible_standing_tax_rate": 0.1,
    }
    if poco_details:
        poco_params.update(poco_details)
    PocoDetails.objects.create(**poco_params)
    return structure


def create_starbase(detail=None, fuels=None, **kwargs):
    if "eve_moon_id" not in kwargs and "eve_moon" not in kwargs:
        kwargs["eve_moon_id"] = 40161465  # Amamake II - Moon 1
    if "eve_type_id" not in kwargs and "eve_type" not in kwargs:
        kwargs["eve_type_id"] = EveTypeId.CALDARI_CONTROL_TOWER.value
    params = {
        "state": Structure.State.POS_ONLINE,
        "fuel_expires_at": now() + dt.timedelta(days=3),
        "position_x": 0,
        "position_y": 0,
        "position_z": 0,
    }
    params.update(kwargs)
    structure = _create_structure(**params)
    assert structure.is_starbase
    detail_params = {
        "allow_alliance_members": True,
        "allow_corporation_members": True,
        "anchor_role": StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        "attack_if_at_war": True,
        "attack_if_other_security_status_dropping": True,
        "fuel_bay_take_role": StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        "fuel_bay_view_role": StarbaseDetail.Role.STARBASE_FUEL_TECHNICIAN_ROLE,
        "last_modified_at": now(),
        "offline_role": StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        "online_role": StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        "structure": structure,
        "unanchor_role": StarbaseDetail.Role.CONFIG_STARBASE_EQUIPMENT_ROLE,
        "use_alliance_standings": True,
    }
    if detail:
        detail_params.update(detail)
    detail_obj = StarbaseDetail.objects.create(**detail_params)
    if not fuels:
        fuels = [
            {"quantity": 960, "eve_type_id": EveTypeId.NITROGEN_FUEL_BLOCK},
            {"quantity": 12000, "eve_type_id": EveTypeId.STRONTIUM},
        ]
    for fuel in fuels:
        StarbaseDetailFuel.objects.create(
            detail=detail_obj,
            eve_type_id=fuel["eve_type_id"],
            quantity=fuel["quantity"],
        )
    return structure


def create_upwell_structure(**kwargs) -> Structure:
    params = {
        "state": Structure.State.SHIELD_VULNERABLE,
        "fuel_expires_at": now() + dt.timedelta(days=3),
    }
    if "eve_type_id" not in kwargs and "eve_type" not in kwargs:
        params["eve_type_id"] = 35832  # Astrahus
    params.update(kwargs)
    structure = _create_structure(**params)
    assert structure.is_upwell_structure
    return structure


def create_jump_gate(**kwargs) -> Structure:
    kwargs["eve_type_id"] = EveTypeId.JUMP_GATE
    if "eve_type" in kwargs:
        del kwargs["eve_type"]
    structure = create_upwell_structure(**kwargs)
    assert structure.is_jump_gate
    return structure


def create_structure_service(**kwargs) -> StructureService:
    if "name" not in kwargs:
        kwargs["name"] = "Clone Bay"
    if "state" not in kwargs:
        kwargs["state"] = StructureService.State.ONLINE
    return StructureService.objects.create(**kwargs)


def create_webhook(**kwargs) -> Webhook:
    while True:
        num = randint(1, 100_000)
        name = f"Test Webhook #{num}"
        if not Webhook.objects.filter(name=name).exists():
            break
    params = {
        "name": name,
        "url": f"https://www.example.com/webhooks/{slugify(name)}",
        "is_active": True,
    }
    params.update(kwargs)
    return Webhook.objects.create(**params)


def _generate_unique_id(Model: object, field_name: str):
    while True:
        id = randint(1, 100_000_000_000)
        params = {field_name: id}
        if not Model.objects.filter(**params).exists():
            return id



"""War embeds."""

# pylint: disable=missing-class-docstring


import dhooks_lite

from django.utils.html import strip_tags
from django.utils.translation import gettext as __
from eveuniverse.models import EveEntity

from app_utils.datetime import ldap_time_2_datetime

from structures.models import Notification, Webhook

from .helpers import (
    gen_eve_entity_link,
    gen_eve_entity_link_from_id,
    target_datetime_formatted,
)
from .main import NotificationBaseEmbed


class NotificationWarEmbed(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._declared_by, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["declaredByID"]
        )
        self._against, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["againstID"]
        )
        self._thumbnail = dhooks_lite.Thumbnail(
            self._declared_by.icon_url(size=self.ICON_DEFAULT_SIZE)
        )


class NotificationCorpWarSurrenderMsg(NotificationWarEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = __("One party has surrendered")
        self._description = __(
            "The war between %(against)s and %(declared_by)s is coming to an end "
            "as one party has surrendered. "
            "The war will be declared as being over after approximately 24 hours."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarAdopted(NotificationWarEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        alliance, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["allianceID"]
        )
        self._title = __("War update: %(against)s has left %(alliance)s") % {
            "against": self._against.name,
            "alliance": alliance.name,
        }
        self._description = __(
            "There has been a development in the war between %(declared_by)s "
            "and %(alliance)s.\n"
            "%(against)s is no longer a member of %(alliance)s, "
            "and therefore a new war between %(declared_by)s and %(against)s has begun."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "alliance": gen_eve_entity_link(alliance),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarDeclared(NotificationWarEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = __("%(declared_by)s Declares War Against %(against)s") % {
            "declared_by": self._declared_by.name,
            "against": self._against.name,
        }
        self._description = __(
            "%(declared_by)s has declared war on %(against)s with %(war_hq)s "
            "as the designated war headquarters.\n"
            "Within %(delay_hours)s hours fighting can legally occur "
            "between those involved."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "war_hq": Webhook.text_bold(strip_tags(self._parsed_text["warHQ"])),
            "delay_hours": Webhook.text_bold(self._parsed_text["delayHours"]),
        }
        self._color = Webhook.Color.DANGER


class NotificationWarInherited(NotificationWarEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        alliance, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["allianceID"]
        )
        opponent, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["opponentID"]
        )
        quitter, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["quitterID"]
        )
        self._title = __("%(alliance)s inherits war against %(opponent)s") % {
            "alliance": alliance.name,
            "opponent": opponent.name,
        }
        self._description = __(
            "%(alliance)s has inherited the war between %(declared_by)s and "
            "%(against)s from newly joined %(quitter)s. "
            "Within **24** hours fighting can legally occur with %(alliance)s."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "alliance": gen_eve_entity_link(alliance),
            "quitter": gen_eve_entity_link(quitter),
        }
        self._color = Webhook.Color.DANGER


class NotificationWarRetractedByConcord(NotificationWarEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = __("CONCORD invalidates war")
        war_ends = ldap_time_2_datetime(self._parsed_text["endDate"])
        self._description = __(
            "The war between %(declared_by)s and %(against)s "
            "has been retracted by CONCORD.\n"
            "After %(end_date)s CONCORD will again respond to any hostilities "
            "between those involved with full force."
        ) % {
            "declared_by": gen_eve_entity_link(self._declared_by),
            "against": gen_eve_entity_link(self._against),
            "end_date": target_datetime_formatted(war_ends),
        }
        self._color = Webhook.Color.WARNING


class NotificationWarCorporationBecameEligible(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = __(
            "Corporation or alliance is now eligible for formal war declarations"
        )
        self._description = __(
            "Your corporation or alliance is **now eligible** to participate in "
            "formal war declarations. This could be because your corporation "
            "and/or one of the corporations in your alliance owns a structure "
            "deployed in space."
        )
        self._color = Webhook.Color.WARNING


class NotificationWarCorporationNoLongerEligible(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = __(
            "Corporation or alliance is no longer eligible for formal war declarations"
        )
        self._description = __(
            "Your corporation or alliance is **no longer eligible** to participate "
            "in formal war declarations.\n"
            "Neither your corporation nor any of the corporations "
            "in your alliance own a structure deployed in space at this time. "
            "If your corporation or alliance is currently involved in a formal war, "
            "that war will end in 24 hours."
        )
        self._color = Webhook.Color.INFO


class NotificationWarSurrenderOfferMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        isk_value = self._parsed_text.get("iskValue", 0)
        owner_1, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text.get("ownerID1")
        )
        owner_1_link = gen_eve_entity_link(owner_1)
        owner_2_link = gen_eve_entity_link_from_id(self._parsed_text.get("ownerID2"))
        self._title = __("%s has offered a surrender") % (owner_1,)
        self._description = __(
            "%s has offered to end the war with %s in the exchange for %s ISK. "
            "If accepted, the war will end in 24 hours and your organizations will "
            "be unable to declare new wars against each other for the next 2 weeks."
        ) % (owner_1_link, owner_2_link, f"{isk_value:,.2f}")
        self._color = Webhook.Color.INFO


class NotificationAllyJoinedWarMsg(NotificationBaseEmbed):
    def __init__(self, notification: Notification) -> None:
        super().__init__(notification)
        self._title = __("Ally Has Joined a War")
        aggressor, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["aggressorID"]
        )
        ally, _ = EveEntity.objects.get_or_create_esi(id=self._parsed_text["allyID"])
        defender, _ = EveEntity.objects.get_or_create_esi(
            id=self._parsed_text["defenderID"]
        )
        start_time = ldap_time_2_datetime(self._parsed_text["startTime"])
        self._description = __(
            "%(ally)s has joined %(defender)s in a war against %(aggressor)s. "
            "Their participation in the war will start at %(start_time)s."
        ) % {
            "aggressor": gen_eve_entity_link(aggressor),
            "ally": gen_eve_entity_link(ally),
            "defender": gen_eve_entity_link(defender),
            "start_time": target_datetime_formatted(start_time),
        }
        self._thumbnail = dhooks_lite.Thumbnail(
            ally.icon_url(size=self.ICON_DEFAULT_SIZE)
        )
        self._color = Webhook.Color.WARNING
